"""

For live connected tests you must setup a test environment and give
the deployer the following.

  - JUJU_API = API Endpoint for an environment (machine 0 port 17070). Example
        JUJU_API=wss://instance-address.com:17070

  - JUJU_HOME = If your using a non standard juju home location

  - Provider api credentials in their default environment locations. For some
    commands we need to execute the cli at the moment (terminate-machine,
    deploy local charm).

  - JUJU_BIN = If you want to use a non-PATH juju binary


For devs, there is an isolated environment test example below that
will self generate an environment.
"""
import base64
import logging
import os
import shutil
import StringIO
import subprocess
import sys
import tempfile
import time
import unittest

from deployer import (
    ConfigStack,
    Service,
    Charm,
    Deployment,
    Environment,
    Importer,
    ErrorExit)

from deployer import (
    path_join, yaml_dump, yaml_load, setup_logging, dict_merge)


class Base(unittest.TestCase):

    def capture_logging(self, name="", level=logging.INFO,
                        log_file=None, formatter=None):
        if log_file is None:
            log_file = StringIO.StringIO()
        log_handler = logging.StreamHandler(log_file)
        if formatter:
            log_handler.setFormatter(formatter)
        logger = logging.getLogger(name)
        logger.addHandler(log_handler)
        old_logger_level = logger.level
        logger.setLevel(level)

        @self.addCleanup
        def reset_logging():
            logger.removeHandler(log_handler)
            logger.setLevel(old_logger_level)
        return log_file

    def mkdir(self):
        d = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, d)
        return d

    def change_environment(self, **kw):
        """
        """
        original_environ = dict(os.environ)

        @self.addCleanup
        def cleanup_env():
            os.environ.clear()
            os.environ.update(original_environ)

        os.environ.update(kw)


class UtilTests(Base):

    def test_relation_list_merge(self):
        self.assertEqual(
            dict_merge(
                {'relations': [['m1', 'x1']]},
                {'relations': [['m2', 'x2']]}),
            {'relations': [['m1', 'x1'], ['m2', 'x2']]})

    def test_no_rels_in_target(self):
        self.assertEqual(
            dict_merge(
                {'a': 1},
                {'relations': [['m1', 'x1'], ['m2', 'x2']]}),
            {'a': 1, 'relations': [['m1', 'x1'], ['m2', 'x2']]})


class CharmTest(Base):

    def setUp(self):
        d = self.mkdir()
        self.series_path = os.path.join(d, "precise")
        os.mkdir(self.series_path)

        self.charm_data = {
            "charm": "couchdb",
            "build": None,
            "branch": "lp:charms/precise/couchdb",
            "rev": None,
            "charm_url": None,
        }
        self.output = self.capture_logging(
            "deployer.charm", level=logging.DEBUG)

    def test_charm(self):
        params = dict(self.charm_data)
        charm = Charm.from_service("scratch", self.series_path, params)

        charm.fetch()
        self.assertEqual(charm.metadata['name'],  'couchdb')

        charm.rev = 7
        charm.update()
        output = subprocess.check_output(
            ["bzr", "revno", "--tree"], cwd=charm.path)
        self.assertEqual(output.strip(), str(7))

        self.assertFalse(charm.is_modified())
        with open(os.path.join(charm.path, 'revision'), 'w') as fh:
            fh.write('0')
        self.assertTrue(charm.is_modified())

    def test_charm_error(self):
        params = dict(self.charm_data)
        params['branch'] = "lp:charms/precise/zebramoon"
        charm = Charm.from_service("scratch", self.series_path, params)
        self.assertRaises(ErrorExit, charm.fetch)
        self.assertIn('bzr: ERROR: Not a branch: ', self.output.getvalue())


class ConfigTest(Base):

    def setUp(self):
        self.output = self.capture_logging(
            "deployer.config", level=logging.DEBUG)

    def test_config_basic(self):
        config = ConfigStack(['configs/ostack-testing-sample.cfg'])
        config.load()
        self.assertEqual(
            config.keys(),
            [u'openstack-precise-ec2',
             u'openstack-precise-ec2-trunk',
             u'openstack-ubuntu-testing'])
        self.assertRaises(ErrorExit, config.get, 'zeeland')
        result = config.get("openstack-precise-ec2")
        self.assertTrue(isinstance(result, Deployment))

    def test_config(self):
        config = ConfigStack([
            "test_data/stack-default.cfg", "test_data/stack-inherits.cfg"])
        config.load()
        self.assertEqual(
            config.keys(),
            [u'my-files-frontend-dev', u'wordpress'])
        deployment = config.get("wordpress")
        self.assertTrue(deployment)


class ServiceTest(Base):

    def test_service(self):
        data = {
            'branch': 'lp:precise/mysql'}

        s = Service('db', data)
        self.assertEqual(s.name, "db")
        self.assertEqual(s.num_units, 1)
        self.assertEqual(s.constraints, None)
        self.assertEqual(s.config, None)

        data = {
            'branch': 'lp:precise/mysql',
            'constraints': "instance-type=m1.small",
            'options': {"services": "include-file://stack-include.yaml"},
            'num_units': 10}
        s = Service('db', data)
        self.assertEquals(s.num_units, 10)
        self.assertEquals(s.constraints, "instance-type=m1.small")
        self.assertEquals(s.config, {"services": "include-file://stack-include.yaml"})


class DeploymentTest(Base):
    def setUp(self):
        self.output = setup_logging(
            debug=True, verbose=True, stream=StringIO.StringIO())

    def test_deployer(self):
        d = ConfigStack(["test_data/blog.yaml"]).get('wordpress-prod')
        services = d.get_services()
        self.assertTrue([s for s in services if s.name == "newrelic"])

        # Ensure inheritance order reflects reality, instead of merge value.
        self.assertEqual(
            d.data['inherits'],  ['wordpress-stage', 'wordpress-base', 'metrics-base'])

        # Fetch charms to verify late binding config values & validation.
        t = self.mkdir()
        os.mkdir(os.path.join(t, "precise"))
        d.repo_path = t
        d.fetch_charms()

        # Load up overrides and resolves
        d.load_overrides(["key=abc"])
        d.resolve_config()

        # Verify include-base64
        self.assertEqual(d.get_service('newrelic').config, {'key': 'abc'})
        self.assertEqual(
            base64.b64decode(d.get_service('blog').config['wp-content']),
            "HelloWorld")

        # TODO verify include-file

        # Verify relations
        self.assertEqual(
            list(d.get_relations()),
            [('blog', 'db'), ('blog', 'cache'), ('blog', 'haproxy')])


@unittest.skipIf(
    not bool(os.environ.get("JUJU_API")),
    "Test env endpoint must be defined: JUJU_API")
class LiveEnvironmentTest(Base):

    def setUp(self):
        self.endpoint = os.environ.get("JUJU_API")
        self.output = self.capture_logging(
            "deployer.env", log_file=sys.stderr, level=logging.DEBUG)
        self.env = Environment("", self.endpoint)
        self.env.connect()

    def tearDown(self):
        self.env.reset(timeout=240)
        self.env.close()

    def test_env(self):
        # Destroy everything.. consistent baseline
        self.env.reset(timeout=240)

        status = self.env.status()
        self.assertFalse('services' in status)
        self.env.deploy("test-blog", "cs:precise/wordpress")
        self.env.deploy("test-db", "cs:precise/mysql")
        self.env.add_relation("test-db", "test-blog")
        self.env.add_units('test-blog', 1)
        self.env.wait_for_units(timeout=600)

        status = self.env.status()
        services = ["test-blog", "test-db"]
        self.assertEqual(
            sorted(status['services'].keys()),
            services)
        for s in services:
            for u in status['services'][s]['units']:
                self.assertEqual(u['agent-state'], "started")


@unittest.skipIf(
    not bool(os.environ.get("JUJU_NEW_ENV")),
    "Skipping new env tests unless new env specified (JUJU_NEW_ENV)")
class DISABLEDIsolatedEnvironmentTest(Base):
    """
    Environments tests that *create* & destroy isolated environments for testing.
    """

    @classmethod
    def _exec(cls, params, ign_err=False, **env_vars):
        env = dict(os.environ)
        env.update(env_vars)
        env["JUJU_HOME"] = cls.cls_juju_home

        print "cmd:", " ".join(params), "env:", " ".join(
            ["%s=%s" % (k, env[k]) for k in env if k.startswith("JUJU")])

        try:
            return subprocess.check_output(
                params, stderr=subprocess.STDOUT, env=env)
        except subprocess.CalledProcessError, e:
            print "Error on exec"
            print " ".join(params), env["JUJU_HOME"]
            print e.output

    @classmethod
    def setUpClass(cls):
        # Bootstraps a new juju test environment
        cls.cls_juju_home = tempfile.mkdtemp()
        cls.juju_bin = os.environ.get("JUJU_BIN", "juju")

        # Create a new environment
        conf = yaml_load(cls._exec([cls.juju_bin, 'init']))
        envs = conf['environments']
        env_type = os.environ.get("JUJU_ENV", "amazon")

        assert env_type in envs, "Invalid env type specified %s" % (envs.keys())
        conf["default"] = env_type

        # Find a public key to use for the environment
        found = False
        key_set = ["~/.ssh/id_dsa.pub", "~/.ssh/id_rsa.pub"]
        for key in key_set:
            key = os.path.expanduser(key)
            if os.path.exists(key):
                found = True
                break
        if not found:
            raise RuntimeError("No ssh public key found: tried (%s" % (
                " ".join(key_set)))
        envs[env_type]['authorized-keys-path'] = key

        # Bootstrap
        with open(path_join(
                cls.cls_juju_home, "environments.yaml"), "w") as fh:
            fh.write(yaml_dump(conf))
        cls._exec([cls.juju_bin, "bootstrap"])

    @classmethod
    def tearDownClass(cls):
        cls._exec([cls.juju_bin, "destroy-environment"])

    def setUp(self):
        self.st = time.time()
        self.change_environment(JUJU_HOME=self.__class__.cls_juju_home)
        self.capture_logging("deployer.env", log_file=sys.stderr, level=logging.DEBUG)

    def test_environment(self):
        env = Environment("")
        st = time.time()
        env.connect()
        print "Bootstrap took", st - time.time()
        status = env.status()
        self.assertFalse(status['services'].keys())


_marker = object()


class Options(dict):

    def __getattr__(self, key):
        v = self.get(key, _marker)
        if v is _marker:
            raise AttributeError(key)
        return v


@unittest.skipIf(
    not bool(os.environ.get("JUJU_API")),
    "Test environment endpoint must be defined JUJU_API")
class ImporterTest(Base):

    def setUp(self):
        self.env = Environment("", os.environ["JUJU_API"])
        self.output = StringIO()
        setup_logging(verbose=True, debug=True, stream=self.output)

        self.importer = Importer()
        self.options = Options({
            'verbose': True, 'watch': True})

        self.importer = Importer(self.env, self.deployment, self.options)

    def get_importer(self, data, dirs=(), **kw):
        pass

    def test_importer(self):
        pass


def collector():
    # import __main__ triggers code re-execution
    from unittest.loader import defaultTestLoader
    __main__ = sys.modules['__main__']
    setupDir = os.path.abspath(os.path.dirname(__main__.__file__))
    return defaultTestLoader.discover(setupDir)


def main():
    unittest.main()

if __name__ == '__main__':
    main()
